import attr
from typing import List
from .event import Event


@attr.s
class EventListWithCount:
    """
    :param result : the list of events
    :param count : the number of events in result
    """
    count = attr.ib(type=int)
    result = attr.ib(type=List[Event], default=attr.Factory(list))



import asyncio
import os
import logging
import yaml
from .app import create_app
from .models.config import Config


def from_yaml(path):
    with open(path) as fh:
        return Config(yaml.load(fh.read()))

ROOT = os.path.dirname(os.path.dirname(__file__))
DEFAULT_CONFIG_PATH = os.path.join(ROOT, "config", "current", "config.yaml")
config_path = DEFAULT_CONFIG_PATH
config = from_yaml(config_path)
app = create_app(config)



import motor.motor_asyncio
import logging
from pymongo.read_preferences import ReadPreference
from .event import Event

LOG = logging.getLogger(__name__)


DB_CLASSES = [
    Event
]


def init_db(db_config):
    db = get_db(db_config)
    map_db_collection_with_class(db, DB_CLASSES)
    return db


def get_db(db_config):
    """
    creates connection with dabatase and gets db object to connect.
    """
    connection = motor.motor_asyncio.AsyncIOMotorClient(
        host=db_config.uri,
        # secondarypreferred has proved to be a good practice,
        # as it enabled minimal contention on the node that should
        # be critical and handle writes. Oftentimes delay of event
        # propagation is in the milliseconds.
        readPreference="secondaryPreferred",
    )
    db = connection[db_config.db_name]
    return db


def map_db_collection_with_class(db, db_classes):
    """
    maps database-collection with classes
    """
    for db_class in db_classes:
        name = db_class._collection
        if name:
            setattr(db, name, db_class(getattr(db, name)))



import datetime
import os
import pytest
import socket
from unittest import mock
from datetime import timedelta
from aiohttp import request
import asyncio
from tycho.app import create_app
from tycho.db import init_db
from tycho.models.config import Config, Mongo
from tycho.models.eventdb import EventDB as ModelEventDB
from tycho.models.event import Event as ModelEvent, CATTRS_CONVERTER
from tycho.routes.event import Event

global_source_id = "222f1f77bcf86cd799439011"
# pytest_plugins = 'aiohttp.pytest_plugin'

DB_NAME = "tycho-tests-{}".format(os.getpid())

DB_CONFIG = Mongo({
    "uri": "mongodb://localhost:27017/?maxPoolSize=10&w=1",
    "db_name": DB_NAME,
})


@pytest.fixture
def event_db():
    return mock.Mock()


@pytest.fixture
def collections():
    return [Event]


@pytest.fixture
def config():
    config_json = Config.get_mock_object().to_primitive()
    config_json["mongo"] = DB_CONFIG.to_primitive()
    config_model = Config(config_json)
    return config_model


@pytest.fixture
def app(loop, config, db):
    return create_app(config, db=db)

@pytest.yield_fixture
def db(loop, config):
    asyncio.set_event_loop(loop)
    _db = init_db(config.mongo)
    yield _db
    if _db:
        loop.run_until_complete(_db.command("dropDatabase"))


@pytest.fixture
def cli(loop, aiohttp_client, config, app):
    return loop.run_until_complete(aiohttp_client(app))


@pytest.fixture
def patch_update_time(update_time, monkeypatch):
    with mock.patch("tycho.db.event.serialize.datetime") as mock_datetime:
        mock_datetime.utcnow.return_value = update_time
        yield


@pytest.fixture
def start_time():
    return datetime.datetime.utcnow()


@pytest.fixture
def update_time():
    return datetime.datetime.utcnow()


@pytest.fixture
def end_time(start_time):
    return start_time + timedelta(days=1)


@pytest.fixture
def event(start_time, end_time):
    """
    * should occur before all events.
    """
    global global_source_id
    return ModelEvent(**{
        "id": "5498d53c5f2d60095267a0bb",
        "source_id": global_source_id,
        "parent_id": "111f1f77bcf86cd799439011",
        "start_time": start_time,
        "end_time": end_time,
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "tags": {
            "source": ["deploy"],
            "type": ["deploy/deploy_all"],
            "author": ["user@example.com"],
            "environment": ["monitor_candidate", "monitor_live"],
            "services": ["tycho"],
            "status": ["success"]
        }
    })


@pytest.fixture
def parent_event(end_time):
    """
    should occur right after event
    """
    global global_source_id
    # parent of the event fixture and child of the source
    return ModelEvent(**{
        "id": "111f1f77bcf86cd799439011",
        "source_id": global_source_id,
        "parent_id": "222f1f77bcf86cd799439011",
        "start_time": end_time + datetime.timedelta(hours=1),
        "end_time": end_time + datetime.timedelta(hours=2),
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "tags": {
            "source": ["deploy"],
            "type": ["deploy/deploy_all"],
            "author": ["user@example.com"],
            "environment": ["monitor_candidate", "monitor_live"],
            "services": ["tycho"],
            "status": ["success"]
        }
    })


@pytest.fixture
def child_event_of_source(end_time):
    """
    * should come right after source_event.
    """
    global global_source_id
    # second child of the source event
    return ModelEvent(**{
        "id": "333f1f77bcf86cd799439333",
        "source_id": global_source_id,
        "parent_id": global_source_id,
        "start_time": end_time + datetime.timedelta(hours=5),
        "end_time": end_time + datetime.timedelta(hours=6),
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "tags": {
            "source": ["deploy"],
            "type": ["deploy/deploy_all"],
            "author": ["user@example.com"],
            "environment": ["monitor_candidate", "monitor_live"],
            "services": ["tycho"],
            "status": ["success"]
        }
    })


@pytest.fixture
def source_event(end_time):
    """
    * should come after parent_event
    """
    return ModelEvent(**{
        "id": "222f1f77bcf86cd799439011",
        "start_time": end_time + datetime.timedelta(hours=3),
        "end_time": end_time + datetime.timedelta(hours=4),
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "tags": {
            "source": ["deploy"],
            "type": ["deploy/deploy_all"],
            "author": ["user@example.com"],
            "environment": ["monitor_candidate", "monitor_live"],
            "services": ["tycho"],
            "status": ["success"]
        }
    })


@pytest.fixture
def update_event(start_time, end_time):
    return ModelEvent(**{
        "id": "578524407e7ac034e407dbd2",
        "start_time": start_time - datetime.timedelta(days=3),
        "end_time": end_time - datetime.timedelta(days=2),
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "tags": {
            "source": ["deploy"],
            "type": ["deploy/deploy_all"],
            "author": ["user@example.com"],
            "environment": ["monitor_candidate", "monitor_live"],
            "services": ["tycho"],
            "status": ["success"]
        }
    })


@pytest.fixture
def event_in_db(loop, event, db, parent_event_in_db):
    loop.run_until_complete(db.event.save(event))
    return event


@pytest.fixture
def parent_event_in_db(loop, parent_event, db, source_event_in_db):
    loop.run_until_complete(db.event.save(parent_event))
    return parent_event


@pytest.fixture
def source_event_in_db(loop, source_event, db, patch_update_time):
    loop.run_until_complete(db.event.save(source_event))
    return source_event


@pytest.fixture
def child_event_of_source_in_db(loop, child_event_of_source, db):
    loop.run_until_complete(db.event.save(child_event_of_source))
    return child_event_of_source


@pytest.fixture
def update_event_in_db(loop, update_event, db, patch_update_time):
    loop.run_until_complete(db.event.save(update_event))
    return update_event


@pytest.fixture
def eventdb(start_time, end_time, update_time):
    cfg = {
        "id": "5498d53c5f2d60095267a0bb",
        "detail_urls": {"jira": "http://jira", "graphite": "http://graphite"},
        "description": "This is a trigger_deploy event.",
        "time": [start_time, end_time],
        "update_time": update_time,
        "tags": [
            "parent_id:111f1f77bcf86cd799439011",
            "source_id:222f1f77bcf86cd799439011",
            "author:user@example.com",
            "environment:monitor_candidate",
            "environment:monitor_live",
            "services:tycho",
            "source:deploy",
            "status:success",
            "type:deploy/deploy_all"
          ]
        }
    return CATTRS_CONVERTER.structure(cfg, ModelEventDB)



from ...models.eventdb import EventDB
from ...models.event import Event
from typing import Dict, List


_reserved_fields_in_eventdb_tag = {
    "source_id", "parent_id"
}


def deserialize_db_event(eventdb: Dict) -> Event:
    """
    Transforms DB event format to Public event format
    """
    new_event = {}

    if eventdb:

        if "tags" in eventdb:
            new_event.update(_extract_tags(eventdb["tags"]))

        if "time" in eventdb:
            new_event.update(_extract_time(eventdb["time"]))

        if "_id" in eventdb:
            new_event["id"] = str(eventdb["_id"])

        if "detail_urls" in eventdb:
            new_event["detail_urls"] = eventdb["detail_urls"]

        if "description" in eventdb:
            new_event["description"] = eventdb["description"]

    return Event(**new_event)


def _extract_tags(tags: Dict) -> Dict:
    """
    Transforms list of key-value pair as strings to a
    dictionary of key-value pairs.
    """
    new_event = {}
    tagged_fields = {}

    if tags:
        for doc in tags:
            key, value = doc.split(':', 1)
            # not a tag key
            if key in _reserved_fields_in_eventdb_tag:
                new_event[key] = value
            # key already present, no need to initialize.
            else:
                if key in tagged_fields:
                    tagged_fields[key].append(value)
                else:
                    tagged_fields[key] = [value]

    if len(tagged_fields) > 0:
        new_event["tags"] = tagged_fields

    return new_event


def _extract_time(time: List) -> Dict:
    """
    Transforms array of times to dictionary of start and end time.
    """

    if not time:
        return {}
    elif len(time) > 2:
        raise ValueError("should not have 3 times : {time}".format(time=time))
    elif len(time) == 2:
        return {"start_time": time[0], "end_time": time[1]}
    elif len(time) == 1:
        return {"start_time": time[0], "end_time": time[0]}



class async_generator:
    '''
        We dont want application layer to directly interact
        with Mongo's cursor. This class acts as middle man
        between Mongo and app codeby encapsulating Mongo's
        cursor and providing iterator to client code.
        Use 'async for' to iterate on the returned object.
    '''

    def __init__(self, cursor, map):
        self.cursor = cursor
        self.map = map

    def __aiter__(self):
        return self

    async def __anext__(self):
        async for doc in self.cursor:
            return self.map(doc)
        else:
            raise StopAsyncIteration



import os
import logging
from aiohttp import web
from aiohttp_transmute import TransmuteUrlDispatcher
from .routes import add_routes
from .db import init_db as init_db
from orbital_core import bootstrap_app


APP_ROOT = os.path.dirname(__file__)


def create_app(config, **kwargs):
    app = web.Application()
    bootstrap_app(app, APP_ROOT,
                  service_name="tycho",
                  service_description="A service for tracking operational change.")
    add_routes(app)
    # add attributes
    app["config"] = config
    app.update(**kwargs)
    app.on_startup.append(lambda app: init_app(app, config))
    return app


async def init_app(app, config):
    if "db" not in app:
        app["db"] = init_db(config.mongo)  # pragma: no cover



import pymongo
from ..utils import async_generator
from ...models.eventnode import EventNode
from ...models.event import Event as ModelEvent
from aiohttp.web import HTTPNotFound
import queue as q

from .deserialize import deserialize_db_event
from .serialize import serialize_to_db_event

MAX_RECURSIVE_DEPTH = 100


class Event:
    _collection = 'event'

    indexes = [
        {'unique': False, 'keys': [("tags", pymongo.ASCENDING)]},
        {'unique': False, 'keys': [("time", pymongo.ASCENDING)]},
        {'unique': False, 'keys': [("update_time", pymongo.ASCENDING)]}
    ]

    def __init__(self, collection):
        self.collection = collection

    async def save(self, data: ModelEvent):
        new_db_format = serialize_to_db_event(data)
        result = await self.collection.insert_one(new_db_format)
        return result

    async def find_one(self) -> ModelEvent:
        result = await self.collection.find_one()
        return deserialize_db_event(result)

    async def find_by_id(self, id) -> ModelEvent:
        document = None
        cursor = self.collection.find({"_id": id})
        while(await cursor.fetch_next):
            document = cursor.next_object()
        if document is None:
            raise HTTPNotFound(text="cannot find event {0}".format(id))
        return deserialize_db_event(document)

    async def update_by_id(self, id, update_doc: ModelEvent, insert: bool = False):
        new_data = serialize_to_db_event(update_doc)
        result = await self.collection.replace_one(
            {"_id": id}, new_data, upsert=insert)
        return result

    async def find_by_parent_id(self, id):
        cursor = self.collection.find({"tags":
                                       {"$in": ["parent_id:{0}".format(id)]}})
        return async_generator(cursor, deserialize_db_event)

    async def find(self, tags=None, frm=None, to=None, use_update_time=False,
                   count=100, page=1):
        if count < 0:
            raise ValueError("Count must be greater than or equal to zero.")

        if page < 1:
            raise ValueError(
                "Page count must be greater than or equal to one.")

        time_field = "update_time" if use_update_time else "time"

        query = {}
        if tags is not None:
            query["tags"] = {"$all": tags}

        # optimize common use case
        # https://docs.mongodb.com/manual/core/multikey-index-bounds/
        if frm and to and not use_update_time:
            query[time_field] = {
                "$elemMatch": {
                    "$gte": frm,
                    "$lt": to,
                }
            }
        else:
            if frm is not None:
                query[time_field] = {"$gte": frm}
            if to is not None:
                if query.get(time_field) is None:
                    query[time_field] = {"$lt": to}
                else:
                    query[time_field]["$lt"] = to

        cursor = self.collection.find(query).sort([(time_field, -1)]).skip(
            (page-1)*count).limit(count)
        return async_generator(cursor, deserialize_db_event)

    async def get_tree(self, id):
        root_event = await self.find_by_id(id)
        root_event_node = EventNode(event=root_event)
        queue = q.Queue()
        queue.put(root_event_node)
        while (not queue.empty()):
            event_node = queue.get()
            children = await self.find_by_parent_id(event_node.event.id)
            async for child in children:
                child_node = EventNode(event=child)
                if child_node not in event_node.children:
                    event_node.children.append(child_node)
                queue.put(child_node)
        return root_event_node

    async def trace(self, event_id):
        """ return back the root-level parent id of the event """
        result = []

        currId = event_id
        curr_depth = 0
        while currId and curr_depth < MAX_RECURSIVE_DEPTH:
            try:
                doc = await self.find_by_id(currId)
                result.append(doc)
                currId = doc.parent_id
                curr_depth += 1
            except HTTPNotFound:
                currId = None
        return result

    async def delete_by_id(self, id) -> bool:
        """ deletes event with provided id and returns True otherwise False"""
        result_map = {0: False, 1: True, None: False}
        result = await self.collection.delete_one({"_id": id})
        return result_map[result.deleted_count]



import attr
from datetime import datetime, timedelta
import json
from unittest.mock import patch
import pytest

from aiohttp.web import HTTPNotFound
from tycho.routes.event import _merge, _update
from tycho.models.event import Event
from schematics.exceptions import DataError


async def test_get_event(event_in_db, cli):
    resp = await cli.get('/api/v1/event/{0}'.format(
        event_in_db.id
    ))
    retrieved_event_json = (await resp.json())
    retrieved_event = Event.from_dict(retrieved_event_json)
    assert (retrieved_event.to_primitive()
            == event_in_db.to_primitive())
    assert resp.status == 200


async def test_get_children_events(event_in_db, cli):
    resp = await cli.get('/api/v1/event/{0}/children'.format(
        event_in_db.parent_id
    ))
    retrieved_event_json = (await resp.json())
    retrieved_event = Event.from_dict(retrieved_event_json[0])
    assert (retrieved_event.to_primitive()
            == event_in_db.to_primitive())
    assert resp.status == 200


async def test_get_event_with_no_query_parameters(source_event_in_db,
                                                  parent_event_in_db,
                                                  event_in_db,
                                                  cli):
    resp = await cli.get('/api/v1/event/')
    assert resp.status == 200

    retrieved_event_json = (await resp.json())
    assert retrieved_event_json["result"][0] == \
        source_event_in_db.to_primitive()
    assert retrieved_event_json["result"][2] == \
        event_in_db.to_primitive()
    assert retrieved_event_json["count"] == 3


async def test_get_event_with_page_parameters(source_event_in_db,
                                              parent_event_in_db,
                                              event_in_db,
                                              cli):
    resp = await cli.get('/api/v1/event/?page=1')
    assert resp.status == 200

    retrieved_event_json = (await resp.json())
    assert retrieved_event_json["result"][0] == \
        source_event_in_db.to_primitive()
    assert retrieved_event_json["result"][2] == \
        event_in_db.to_primitive()
    assert retrieved_event_json["count"] == 3

    # page 2
    resp = await cli.get('/api/v1/event/?page=2')
    assert resp.status == 200

    retrieved_event_json = (await resp.json())
    assert retrieved_event_json["count"] == 0


async def test_get_events_with_timestamp_queries(parent_event_in_db,
                                                 event_in_db,
                                                 cli):
    """
    timestamp queries should return events up, but not including,
    the to. frm should be inclusive.
    """
    frm = event_in_db.start_time.isoformat()
    to = parent_event_in_db.start_time.isoformat()
    params = {"frm": frm, "to": to}
    resp = await cli.get('/api/v1/event/', params=params)
    event_json = (await resp.json())
    assert event_json["count"] == 1
    assert event_in_db.to_primitive() == event_json["result"][0]

    to = (parent_event_in_db.start_time + timedelta(seconds=1)).isoformat()
    params = {"frm": frm, "to": to}
    resp = await cli.get('/api/v1/event/', params=params)
    event_json = (await resp.json())
    assert event_json["count"] == 2
    assert event_in_db.to_primitive() == event_json["result"][1]
    assert parent_event_in_db.to_primitive() == event_json["result"][0]


async def test_get_events_with_update_timestamp_queries(parent_event_in_db,
                                                        event_in_db,
                                                        update_time,
                                                        cli):
    frm = (update_time - timedelta(seconds=3)).isoformat()
    to = (update_time - timedelta(seconds=1)).isoformat()
    params = {"frm": frm, "to": to, "use_update_time": "true"}
    resp = await cli.get('/api/v1/event/', params=params)
    event_json = (await resp.json())
    assert event_json["count"] == 0

    to = (update_time + timedelta(seconds=1)).isoformat()
    params = {"frm": frm, "to": to}
    resp = await cli.get('/api/v1/event/', params=params)
    event_json = (await resp.json())
    assert event_json["count"] == 1
    assert event_in_db.to_primitive() == event_json["result"][0]


@pytest.mark.parametrize("frm, to, res",
                         [('None', 'None', 400),
                          ('', '', 400),
                             ('invalid frm time', 'invalid to time', 400)]
                         )
async def test_get_events_with_invalid_timestamp_queries(frm, to, res,
                                                         cli):
    params = {"frm": frm, "to": to}
    resp = await cli.get('/api/v1/event/', params=params)
    assert resp.status == res


async def test_get_events_with_timestamp_tag_count(source_event_in_db,
                                                   child_event_of_source_in_db,
                                                   parent_event_in_db,
                                                   event_in_db,
                                                   cli):
    frm = source_event_in_db.start_time.isoformat()
    to = child_event_of_source_in_db.start_time.isoformat()
    params = [
        ('frm', frm), ('to', to), ('tag', 'environment:monitor_candidate'),
        ('tag', 'services:tycho'), ('count', "1"),
    ]
    # the find API should be returning back values with an exclusive upper bound.
    # in other words, event_in_db starts at to, but will not be included
    # in the query.
    resp = await cli.get('/api/v1/event/', params=params)
    event_json = (await resp.json())
    assert event_json["count"] == 1
    assert event_json["result"][0] == source_event_in_db.to_primitive()


async def test_get_trace_until_source(event_in_db,
                                      parent_event_in_db,
                                      source_event_in_db,
                                      cli):
    resp = await cli.get('/api/v1/event/{0}/trace'.format(
        event_in_db.id
    ))
    event_json = (await resp.json())
    result = [Event.from_dict(elem) for elem in event_json]
    ls = [event_in_db, parent_event_in_db, source_event_in_db]
    for i in range(3):
        assert ls[i].to_primitive() == result[i].to_primitive()


async def test_get_trace_for_source(source_event_in_db,
                                    cli):
    resp = await cli.get('/api/v1/event/{0}/trace'.format(
        source_event_in_db.id
    ))
    event_json = (await resp.json())
    result = [Event.from_dict(elem) for elem in event_json]
    assert source_event_in_db.to_primitive() == result[0].to_primitive()


async def test_get_trace_parent_have_two_children(parent_event_in_db,
                                                  source_event_in_db,
                                                  child_event_of_source_in_db,
                                                  cli):
    resp = await cli.get('/api/v1/event/{0}/trace'.format(
        parent_event_in_db.id
    ))
    event_json = (await resp.json())
    result = [Event.from_dict(elem).to_primitive() for elem in event_json]
    assert child_event_of_source_in_db.to_primitive() not in result


async def test_get_event_impact(cli,
                                source_event_in_db,
                                child_event_of_source_in_db,
                                parent_event_in_db,
                                event_in_db):
    resp = await cli.get('/api/v1/event/{0}/impact'.format(
        source_event_in_db.id
    ))
    assert resp.status == 200

    source_node = (await resp.json())
    assert source_node["event"] == \
        source_event_in_db.to_primitive()

    first_source_child = source_node["children"][0]["event"]
    assert first_source_child == \
        child_event_of_source_in_db.to_primitive()

    second_source_child = source_node["children"][1]
    assert second_source_child["event"] == \
        parent_event_in_db.to_primitive()

    leaf = second_source_child["children"][0]["event"]
    assert leaf == event_in_db.to_primitive()

    child_of_leaf = second_source_child["children"][0]["children"]
    assert child_of_leaf == []


async def test_get_event_impact_diagram(cli, event_in_db):
    resp = await cli.get('/event/{0}/'.format(
        event_in_db.id
    ))
    assert resp.status == 200
    assert resp.content_type == 'text/html'


async def test_get_event_impact_diagram_with_invalid_id(cli):
    event_id = 'dummy_id'
    resp = await cli.get('/event/{0}/'.format(event_id))
    assert resp.status == 404
    resp_text = await resp.text()
    assert resp_text == 'Event ID {} is not available in DB'.format(event_id)


@pytest.mark.parametrize("log", [False, True])
async def test_put_event(db, event, cli, log, app):
    with patch("tycho.routes.event.LOG") as mock_log:
        app["config"].log_events = log
        resp = await cli.put(f'/api/v1/event/',
                             headers={"content-type": "application/json"},
                             data=json.dumps({"event":
                                              event.to_primitive()}))
        assert resp.status == 200
        retrieved_event_json = (await resp.json())
        retrieved_event = Event.from_dict(retrieved_event_json)
        resp2 = await cli.get('/api/v1/event/{0}'.format(
            event.id
        ))
        assert resp2.status == 200
        retrieved_event2_json = (await resp2.json())
        retrieved_event2 = Event.from_dict(retrieved_event2_json)
        assert (retrieved_event.to_primitive()
                == retrieved_event2.to_primitive())
        if log:
            mock_log.info.assert_called_with(event.to_primitive())
        else:
            assert mock_log.assert_not_called


async def test_put_invalid_event_raises_exception(cli):
    event_json = {'source_id': """
{
    "affected_services":["static-pre"],
    "source":"orchestra",
    "phase":"begin",
    "commithash":"cbc010a240e6fa5c9b57ec3a367cd1a6b3b6",
    "hst":"host2.localhost",
    "committer_email":"user2@example.com",
    "author":"user2@example.com"
}
    """.strip()}

    resp = await cli.put('/api/v1/event/',
                         headers={"content-type": "application/json"},
                         data=json.dumps({"event": event_json})
                         )
    assert resp.status == 400


async def test_post_invalid_event_raises_bad_request(event, cli):
    event_json = {'source_id': """
{
    "affected_services":["static-pre"],
    "source":"orchestra",
    "phase":"begin",
    "commithash":"cbc010a240e6fa5c9b57ec3a367cd1a6b3b6",
    "hst":"host2.localhost",
    "committer_email":"user2@example.com",
    "author":"user2@example.com"
}
    """.strip()}
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation": "merge",
                                           "event": event_json})
                          )
    assert resp.status == 400


async def test_put_invalid_event_raises_exception(cli):
    resp = await cli.put('/api/v1/event/',
                         headers={"content-type": "application/json"},
                         data=json.dumps({"event": {"bad_key": "hello"}})
                         )
    assert resp.status == 400


async def test_post_invalid_event_raises_bad_reqeuest(event, cli):
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation": "merge",
                                           "event": {"bad_key": "hello"}})
                          )
    assert resp.status == 400


async def test_post_dont_allow_invalid_opers(event, cli):
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation":
                                           "not_merge_nor_update",
                                           "event":
                                           event.to_primitive()}))
    assert resp.status == 400


@pytest.mark.parametrize("log", [False, True])
async def test_post_event_merge(event, cli, app, log):
    with patch("tycho.routes.event.LOG") as mock_log:
        app["config"].log_events = log
        resp = await cli.post('/api/v1/event/',
                              headers={"content-type": "application/json"},
                              data=json.dumps({"operation":
                                               "merge",
                                               "event":
                                               event.to_primitive()}))
        retrieved_event_json = (await resp.json())
        retrieved_event = Event.from_dict(retrieved_event_json)
        assert (retrieved_event.to_primitive()
                == event.to_primitive())
        assert resp.status == 200
        if log:
            mock_log.info.assert_called_with(event.to_primitive())
        else:
            assert mock_log.assert_not_called


async def test_post_event_without_id(event, cli):
    new_event = Event(
        {"description": "Event without id",
         "tags": {"author": ["orbital@example.com"]}
         })
    assert new_event.id != None

    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation":
                                           "merge",
                                           "event":
                                           new_event.to_primitive()}))
    assert resp.status == 200


@pytest.mark.parametrize("upsert", [(True), (False)])
async def test_post_existing_event_with_upsert_option(upsert, event_in_db, cli):
    expected_author = event_in_db.tags["author"]
    new_event = {"id": event_in_db.id,
                 "start_time": str(event_in_db.start_time - timedelta(seconds=50)),
                 "tags": {"author": ["orbital@example.com"]}
                 }
    expected_author.extend(["orbital@example.com"])

    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({
                                "operation": "merge",
                                "event": new_event,
                                "insert": upsert,
                            }))
    assert resp.status == 200

    merged_event_json = (await resp.json())
    merged_event = Event.from_dict(merged_event_json)
    assert set(expected_author) == set(merged_event.tags["author"])
    delta = event_in_db.start_time - merged_event.start_time
    assert delta.total_seconds() == 50


async def test_post_new_event_with_upsert_option_true(app, event, cli):
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({
                                     "operation": "merge",
                                     "event": event.to_primitive(),
                                     "insert": True
                                  }))
    assert resp.status == 200
    merged_event_json = (await resp.json())
    merged_event = Event.from_dict(merged_event_json)

    result = await app["db"].event.find_by_id(event.id)
    assert result.to_primitive() == event.to_primitive()


async def test_post_new_event_with_upsert_option_false(app, event, cli):
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({
                                     "operation": "merge",
                                     "event": event.to_primitive(),
                                     "insert": False
                                 }))
    assert resp.status == 200
    merged_event_json = (await resp.json())
    merged_event = Event.from_dict(merged_event_json)

    with pytest.raises(HTTPNotFound):
        await app["db"].event.find_by_id(event.id)


@pytest.mark.parametrize("new_description, should_merge",
                         [(" trigger_deploy ", False),
                          ("This is a new trigger_deploy event", True),
                             ("", False)])
async def test_merge_description_with_post_existing_event(new_description, should_merge, event_in_db, cli):
    new_event = Event.from_dict(
        {"id": event_in_db.id,
         "description": new_description
         })

    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation":
                                           "merge",
                                           "event":
                                           new_event.to_primitive(),
                                           "insert":
                                           True}))
    assert resp.status == 200
    merged_event_json = (await resp.json())
    merged_event = Event.from_dict(merged_event_json)
    assert (merged_event.description
            == event_in_db.description) != should_merge


def test_merge_empty_event_with_new_one():
    existing_event = Event()
    new_event = Event.from_dict({"id": existing_event.id,
                                 "source_id": "5498d53c5f2d60095267a0bc",
                                 "end_time": existing_event.end_time + timedelta(hours=2)})
    _merge(existing_event, new_event)
    assert existing_event.id == new_event.id
    assert existing_event.source_id == new_event.source_id
    assert existing_event.end_time == new_event.end_time
    assert len(attr.asdict(existing_event)) == len(attr.asdict(new_event))


def test_update_empty_event_with_new_one():
    existing_event = Event()
    new_event = Event(source_id="5498d53c5f2d60095267a0bc")
    _update(existing_event, new_event)
    assert attr.asdict(existing_event) == attr.asdict(new_event)


def test_merge_diff_values_of_reserved_keys_raises_exception():
    existing_event = Event({"source_id": "5498d53c5f2d60095267a0bb"})
    new_event = Event(
        {"id": existing_event.id, "source_id": "5498d53c5f2d60095267a0bc"})
    with pytest.raises(ValueError):
        _merge(existing_event, new_event)


def test_merge_two_start_times():
    time = datetime.now()
    existing_event = Event.from_dict({"start_time": time})
    new_event = Event.from_dict(
        {"id": existing_event.id, "start_time": time - timedelta(hours=2)})
    _merge(existing_event, new_event)
    assert existing_event.start_time == new_event.start_time


def test_merge_two_end_times():
    time = datetime.now()
    existing_event = Event.from_dict({"end_time": time})
    new_event = Event.from_dict(
        {"id": existing_event.id, "end_time": time + timedelta(hours=2)})
    _merge(existing_event, new_event)
    assert existing_event.end_time == new_event.end_time


def test_merge_detail_urls():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "detail_urls":
                                      {"graphite": "http://graphite",
                                       "concrete": "http://concrete"}})

    new_event = Event.from_dict({"id": existing_event.id,
                                 "detail_urls": {"graphite": "http://graphite",
                                                 "concrete": "http://concrete"}})
    _merge(existing_event, new_event)
    ls = {"graphite": "http://graphite",
          "concrete": "http://concrete"}
    assert existing_event.detail_urls == ls


def test_merge_with_already_existing_keys():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "tags": {"author": ["Sean"]}})
    new_event = Event.from_dict({"id": existing_event.id,
                                 "source_id": "5498d53c5f2d60095267a0bb",
                                 "tags": {"author": ["Yusuke"]}})
    _merge(existing_event, new_event)
    assert existing_event.tags["author"] == ["Sean", "Yusuke"]


def test_merge_two_descriptions():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "description":
                                      "This is a Concrete Event."})
    new_event = Event.from_dict({"id": existing_event.id,
                                 "description": "The version is 2.2.3."})
    _merge(existing_event, new_event)
    test_str = "This is a Concrete Event.\nThe version is 2.2.3."
    assert existing_event.description == test_str


async def test_post_event_update(event, cli):
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation":
                                           "update",
                                           "event":
                                           event.to_primitive()}))
    retrieved_event_json = (await resp.json())
    retrieved_event = Event.from_dict(retrieved_event_json)
    assert (retrieved_event.to_primitive()
            == event.to_primitive())
    assert resp.status == 200


async def test_post_existing_event_update(event_in_db, cli):
    updated_time = event_in_db.start_time - timedelta(seconds=50)
    new_event = {"id": event_in_db.id,
                 "start_time": str(updated_time),
                 "tags": {"author": ["orbital@example.com"]}
                 }
    resp = await cli.post('/api/v1/event/',
                          headers={"content-type": "application/json"},
                          data=json.dumps({"operation":
                                           "update",
                                           "event":
                                           new_event}))
    assert resp.status == 200
    retrieved_event_json = (await resp.json())
    retrieved_event = Event.from_dict(retrieved_event_json)
    assert event_in_db.id == retrieved_event.id
    assert set(["orbital@example.com"]
               ) == set(retrieved_event.tags["author"])
    assert retrieved_event.start_time == updated_time


async def test_delete_non_existing_event(cli):
    event_id = "event_not_exist"
    resp = await cli.delete("/api/v1/event/{0}/delete".format(event_id))
    assert resp.status == 400
    retrieved_event_json = (await resp.json())
    assert "Can not delete event with id: {0}".format(
        event_id) in retrieved_event_json["result"]


async def test_delete_existing_event(cli, event_in_db):
    resp = await cli.delete("/api/v1/event/{0}/delete".format(event_in_db.id))
    assert resp.status == 200
    retrieved_event_json = (await resp.json())
    assert "Event: {0} deleted successfully".format(
        event_in_db.id) == retrieved_event_json["message"]
    resp = await cli.get('/api/v1/event/{0}'.format(
        event_in_db.id
    ))
    assert 404 == resp.status


def test_same_values_not_added_twice():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "tags": {"author": ["Yusuke"]}})
    new_event = Event.from_dict({"id": existing_event.id,
                                 "tags": {"author": ["Yusuke"]}})
    _merge(existing_event, new_event)
    assert existing_event.tags["author"] == ["Yusuke"]


def test_merge_two_attrs_of_type_lists_with_duplicates():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "tags": {"author": ["Yusuke", "Sean"]}})
    new_event = Event.from_dict({"id": existing_event.id,
                                 "tags": {"author": ["Yusuke", "Mayur"]}})
    _merge(existing_event, new_event)
    assert existing_event.tags["author"] == ["Yusuke", "Sean", "Mayur"]


def test_merge_new_attributes():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "tags": {"author": ["Saroj"]}})
    new_event = Event.from_dict({"id": existing_event.id,
                                 "tags": {"reviewer": ["Yusuke"]}})
    _merge(existing_event, new_event)
    assert existing_event.tags["reviewer"] == ["Yusuke"]


def test_update_non_existing_values():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb"})
    new_event = Event.from_dict({"parent_id": "5498d53c5f2d60095267a0bb"})
    _update(existing_event, new_event)
    assert existing_event.parent_id == "5498d53c5f2d60095267a0bb"


def test_update_existing_values():
    existing_event = Event.from_dict({"source_id": "5498d53c5f2d60095267a0bb",
                                      "tags": {"author": ["Saroj"]}})
    new_event = Event.from_dict({"tags": {"author": ["Yusuke"]}})
    _update(existing_event, new_event)
    assert existing_event.tags["author"] == ["Yusuke"]


async def test_user_route(cli):
    resp = await cli.get('/')
    assert 200 == resp.status
    text = await resp.text()
    assert "html" in text



from .event import (add_event_api, add_statics)
from aiohttp_transmute import add_swagger


def add_routes(app):
    # add apis
    add_event_api(app)
    add_statics(app)
    add_swagger(app, "/api/swagger.json", "/api/")



import pytest
import bson
import datetime
from tycho.models.eventdb import EventDB
from schematics.contrib.mongo import ObjectIdType
from schematics.exceptions import DataError
from schematics.models import Model

from tycho.models.event import CATTRS_CONVERTER


def test_objectidtype_object_empty():
    obj = ObjectIdType()
    with pytest.raises(TypeError):
        obj("507f1f77bcf86cd7994390112")


def test_objectidtype_on_invalid_objectid():
    obj = ObjectIdType()
    with pytest.raises(TypeError):
        obj.validate("123")


def test_objectidtype_on_objectd():
    obj = ObjectIdType()
    try:
        obj.validate(bson.ObjectId("507f1f77bcf86cd799439011"))
    except pytest.raises(TypeError):
        pytest.fail("This exception shouldn't be raised")


def test_objectidtype_errors_at_non_objectid():
    class SubModel(Model):
        _id = ObjectIdType(required=False)
    doc = {"_id": "123"}
    with pytest.raises(DataError):
        # when illegal value present. init raise exception.
        SubModel(doc)


def test_eventdb_no_field_present():
    doc = {}
    # when no illegal value present.
    # below initialization won't raise exception.
    with pytest.raises(TypeError):
        CATTRS_CONVERTER.structure(doc, EventDB)


def test_eventdb_required_fields_not_present():
    doc = {"detail_urls": {"graphite": "http://graphite",
                           "concrete": "http://concrete"}}
    with pytest.raises(TypeError):
        CATTRS_CONVERTER.structure(doc, EventDB)


def test_eventdb_required_fields_present():
    doc = {"id": "507f1f77bcf86cd799439011",
           "update_time": datetime.datetime.utcnow(),
           "time": [datetime.datetime.utcnow()]}
    CATTRS_CONVERTER.structure(doc, EventDB)


def test_eventdb_all_fields_present():
    doc = {"id": "507f1f77bcf86cd799439011",
           "time": [datetime.datetime.utcnow()],
           "update_time": datetime.datetime.utcnow(),
           "detail_urls": {"graphite": "http://graphite"},
           "tags": ["author:yusuket"]}
    CATTRS_CONVERTER.structure(doc, EventDB)



from schematics.models import Model
from schematics.types import StringType, IntType, BooleanType
from schematics.types.compound import ModelType


class Mongo(Model):
    uri = StringType(required=True)
    db_name = StringType(required=True)


class application(Model):
    host = StringType(required=True)
    port = IntType(required=True)


class Config(Model):
    """ a config object that the app accepts. """
    environment = StringType(required=False)
    mongo = ModelType(Mongo, required=True)
    application = ModelType(application, required=True)
    # a configuration parameter to define whether events should be
    # logged any time an event is added/updated
    log_events = BooleanType(required=False, default=True)



import attr
from typing import List
from .event import Event, CATTRS_CONVERTER


@attr.s
class EventNode:
    """
    A class that implements the event traversal
    :param event : the current event
    :param children: its children event as a list
    """
    event = attr.ib(type=Event)
    children = attr.ib(type=List[Event], default=attr.Factory(list))

    def to_primitive(self):
        return CATTRS_CONVERTER.unstructure(self)



from datetime import datetime
from ...models.event import Event
from typing import Dict, List


def serialize_to_db_event(event: Event) -> Dict:
    """
    Transforms public event format to DB event format.
    Ignores key and value with NoneType.
    """

    new_event = {}

    new_event["tags"] = []

    if event.tags is not None:
        new_event["tags"].extend(_get_tags(event.tags))

    for key in ["source_id", "parent_id"]:
        if getattr(event, key):
            new_event["tags"].append(
                "{key}:{value}".format(key=key, value=getattr(event, key)))

    if event.id is not None:
        new_event["_id"] = str(event.id)

    new_event["time"] = []

    for key in ["start_time", "end_time"]:
        if getattr(event, key) is not None:
            new_event["time"].append(getattr(event, key))

    for key in ["detail_urls", "description"]:
        if getattr(event, key) is not None:
            new_event[key] = getattr(event, key)

    new_event["update_time"] = datetime.utcnow()

    return new_event


def _get_tags(event_tags: Dict) -> List:
    """
    Transforms dictionary of key-value pairs (not in untag_fields list)
    to list of key-value pair concatenated as strings inside tags field.
    """
    tags = []

    # event_tags will be an empty dictionary by default.
    # key and value in event_tags will never be of None Type.
    for key in event_tags:
        for value in event_tags[key]:
            tags.append(
             "{key}:{value}".format(key=key, value=value))

    return tags



import aiohttp_transmute
import attr
import json
import os
import logging
from aiohttp import web
from aiohttp.web import HTTPNotFound
from aiohttp_transmute import (APIException, add_route, route)
from datetime import datetime
from ..models.eventnode import EventNode
from ..models.events_with_count import EventListWithCount
from ..models.event import Event

from ..templates import get_template


LOG = logging.getLogger(__name__)


@aiohttp_transmute.describe(methods="GET", paths="/api/v1/event/{event_id}")
async def get_event(request, event_id: str) -> Event:
    """
    returns the event stored in the db given its id
    :param request: the request object
    :param event_id: the unique id corresponding to the event
    :return: Event object
    """
    event = await request.app["db"].event.find_by_id(event_id)
    return event


@aiohttp_transmute.describe(methods="GET",
                            paths="/api/v1/event/{event_id}/children")
async def get_children_events(request, event_id: str) -> [Event]:
    """
     returns a event's children events given the parent's event id
     :param request: the request object
     :param event_id: the unique id corresponding to the parent event
     :return: parent event's children as a list
    """
    result = []
    async for doc in await request.app["db"].event.find_by_parent_id(event_id):
        result.append(doc)
    return result


@aiohttp_transmute.describe(methods="GET",
                            paths="/api/v1/event/{event_id}/trace")
async def get_parent_events_until_source(request, event_id: str) -> [Event]:
    """
     returns the whole branch of events up until the source
     given the child's event id
     :param request: the request object
     :param event_id: the unique id corresponding to the child event
     :return: list of parents until the source as a list
    """
    return (await request.app["db"].event.trace(event_id))


@aiohttp_transmute.describe(methods="GET",
                            paths="/api/v1/event/")
async def get_events(request, count: int = 100,
                     frm: datetime = None, to: datetime = None,
                     use_update_time: bool = False,
                     tag: [str] = None, page: int = 1) -> EventListWithCount:
    """
    return events based on query parameters
    :param request: the client request object

    :param count: the number of events to be returned, default=100

    :param frm: if provided, the events returned will be the ones
            with start_time on and after this timestamp, default=None

    :param to: if provided, the events returned will be the ones
            with start_time before this timestamp, default=None

    :param: use_update_time: search by ets event update time or
             event provided start and end times, default=False

    :param page: the attribute for pagination of events,
               default=1

    :param tag: the list of other query parameters, including
         environment, services and description, default=None

    :return: The 'EventListwithCount' object, which has the list
             of events to be returned in its 'result' field,
             and the number of results in its 'count' field
    """
    result = []
    qry = {"tags": tag, "count": count, "page": page,
           "use_update_time": use_update_time}

    for param in ['frm', 'to']:
        value = eval(param)
        if value:
            qry.update({param: value})

    docs = await request.app["db"].event.find(**qry)
    async for doc in docs:
        result.append(doc)
    count = len(result)
    event_with_count = EventListWithCount(result=result, count=count)
    return event_with_count


@aiohttp_transmute.describe(methods="GET",
                            paths="/api/v1/event/{event_id}/impact")
async def get_event_impact(request, event_id: str) -> EventNode:
    """
    gets the whole impact of an event given its id.
    Eg:

    ```
    1
    | \\
    2  3
    | \\
    4  5
    ```

    Given '1's id, this function would return the node
    representation of the above tree with pointer to the root.

    :param request: the object representing the client request
    :param event_id: the id corresponding to the source
    :return: the impact of the source node.
    """
    event = await request.app["db"].event.get_tree(event_id)
    return event


@aiohttp_transmute.describe(methods="PUT", paths="/api/v1/event/")
async def put_event(request, event: Event) -> Event:
    """
    stores a new event on the database
    :param request: the object representing the client request
    :param event: the event to be stored
    :return: the stored event
    """
    await request.app["db"].event.save(event)
    if request.app["config"].log_events:
        LOG.info(event.to_primitive())
    return event


@aiohttp_transmute.describe(methods="DELETE", paths="/api/v1/event/{event_id}/delete")
async def delete_event(request, event_id: str) -> bool:
    """
        deletes an existing event with event_id
        :param request: the request object
        :param event_id: the id corresponding to the source
        :return: the result of delete operation (msg)
    """
    result = await request.app["db"].event.delete_by_id(event_id)
    if not result:
        raise APIException(
            "Can not delete event with id: {0}".format(event_id))
    return {'message': 'Event: {0} deleted successfully'.format(event_id)}


@aiohttp_transmute.describe(methods="POST", paths="/api/v1/event/")
async def post_event(request, event: Event,
                     operation: str = "merge", insert: bool = False) -> Event:
    """
        merge or update an existing event with a new one
        :param request: the request object
        :param operation: takes either "merge" or "update":
           - merge will merge the existing event with the new event and
             raise an exception whenever merge isn't possible
           - update will update the values in the existing event with
             the new values given by the post request and add new values
        :return: the updated/merged event
    """
    if operation != "merge" and operation != "update":
        raise APIException(
            "only update and merge operation are supported, {0} passed".format(operation))

    try:
        existing_event = await request.app["db"].event.find_by_id(event.id)
        if operation == "merge":
            _merge(existing_event, event)
        else:
            _update(existing_event, event)
    except HTTPNotFound:
        existing_event = event

    if request.app["config"].log_events:
        LOG.info(event.to_primitive())
    await request.app["db"].event.update_by_id(event.id, existing_event, insert)
    return existing_event


def _merge(existing_event, new_event):
    """
    A helper function to merge the existing event with the
    new values given by the new_eventionary

    :param: existing_event: The event that's already in the database
    :param: the dictionary with new attributes to merge with the existing event
    :modifies: existing_event with the merged values
    """
    new_event_dict = attr.asdict(new_event)
    for key in new_event_dict:
        if new_event_dict.get(key):
            # try merging the values which share the same keys
            if getattr(existing_event, key):
                if key != "tags":
                    _merge_reserved_keys(key, existing_event, new_event)
                else:
                    # merge the values so that the resulting value is a dict
                    _add_to_dict(key, existing_event, new_event)
            # else just add the new values to our event
            else:
                setattr(existing_event, key, new_event_dict[key])


def _merge_reserved_keys(key, existing_event, new_event):
    """
    A helper function that merges the reserved keys or raises an exception
    when merge not possible
    :param key: The key to be merged
    :param existing_event: The event that's already in the database
    :param new_event: the new event to be merged
    :return: the dictionary with new attributes to merge with
             the existing event
    """
    # change the start time to be the earlier
    # of the two keys
    if key == "start_time":
        if new_event.start_time is not None:
            existing_event.start_time = min(
                new_event.start_time, existing_event.start_time)

    elif key == "end_time":
        # change the end time to be the later
        # of the two keys
        if new_event.end_time is not None:
            existing_event.end_time = max(
                new_event.end_time, existing_event.end_time)

    elif key == "description":
        # append the new_event's description to the already existing one
        str1 = existing_event.description.strip()
        str2 = new_event.description.strip()
        if str2 and str2 not in str1:
            # Adding new line as a separator of events description
            existing_event.description = str1 + "\n" + str2

    elif key == "detail_urls":
        # merge two detail_urls lists given by new_event and existing_event
        for k in new_event.detail_urls:
            if new_event.detail_urls.get(k) is not None:
                existing_event.detail_urls[k] = new_event.detail_urls[k]
    else:
        # raise exception if the other reserved vals don't match
        if getattr(existing_event, key) != getattr(new_event, key):
            raise ValueError("Merge not possible:"
                             "reserved keys have clashing values.")


def _add_to_dict(key, existing_event, new_event):
    """
    A helper function that merges the keys whose values is a list
    :param key: the key corresponding to the values to be added
    :param existing_event: the already existing event in the db
    :param new_event: the new db to be merged
    :modifies the existing_event with the merged values
    """
    prev = getattr(existing_event, key)
    new = getattr(new_event, key)

    for k in new:
        if k not in prev:
            prev[k] = new[k]
        else:
            for elem in new[k]:
                if elem not in prev[k]:
                    prev[k].append(elem)

    setattr(existing_event, key, prev)


def _update(existing_event, new_event):
    """
    A helper function which updates the existing event in the database
    to new values given by the data in put request

    :param: existing_event: The event that's already in the database
    :param: the dictionary with new attributes to merge with the existing event
    :modifies: existing_event with the merged values

    """
    new_event_dict = attr.asdict(new_event)

    for key in new_event_dict:
        if new_event_dict.get(key) is not None:
            setattr(existing_event, key, new_event_dict[key])


async def get_event_root_and_impact_diagram(request):
    """
    traverses the given event_id to the root, then finds the impact
    of that event.
    """
    event_id = request.match_info['event_id']
    event_tree = (await request.app["db"].event.trace(event_id))
    if not event_tree:
        raise HTTPNotFound(
            text='Event ID {} is not available in DB'.format(event_id))
    root_event = event_tree[-1]
    event = await request.app["db"].event.get_tree(root_event.id)

    body = get_template("diagrams.html").render(
        config=request.app["config"], rawData=json.dumps(event.to_primitive())
    ).encode("UTF-8")

    return web.Response(body=body, content_type="text/html")


def add_statics(app):
    APP_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    STATIC_ROOT = os.path.join(APP_ROOT, "tycho", "static")
    app.router.add_static('/static', STATIC_ROOT)


def add_event_api(app):
    route(app, get_event)
    route(app, post_event)
    route(app, delete_event)
    route(app, put_event)
    route(app, get_children_events)
    route(app, get_parent_events_until_source)
    route(app, get_events)
    route(app, get_event_impact)
    app.router.add_route(
        'GET', "/event/{event_id}/", get_event_root_and_impact_diagram)



from schematics.types import UTCDateTimeType


class UTCDateTimeTypeIgnoreMicroseconds(UTCDateTimeType):

    """
       A variant of ``UTCDateTimeTypeExt`` that stores time
       granularity in milliseconds instead of microseconds.
    """

    def __init__(self, formats=None, **kwargs):
        super(UTCDateTimeType, self).__init__(
            formats=formats, convert_tz=True, drop_tzinfo=True, **kwargs
        )

    def _ignore_microseconds(self, value):
        microsecond = value.microsecond
        new_microsecond = int(microsecond / 1000) * 1000
        return value.replace(microsecond=new_microsecond)

    def to_primitive(self, value, context=None):
        new_dt = self._ignore_microseconds(value)
        return super(UTCDateTimeType, self).to_primitive(new_dt, context)

    def to_native(self, value, context=None):
        old_dt = super(UTCDateTimeType, self).to_native(value, context)
        return self._ignore_microseconds(old_dt)

    def from_string(self, value):
        old_dt = super(UTCDateTimeType, self).from_string(value)
        if old_dt is None:
            return None
        else:
            return self._ignore_microseconds(old_dt)

    def _mock(self, context=None):
        old_dt = super(UTCDateTimeType, self)._mock(context)
        return self._ignore_microseconds(old_dt)

    def pre_setattr(self, value):
        return self._ignore_microseconds(value)    # pragma: no cover



import os
import sys
from setuptools import setup, find_packages

BASE = os.path.dirname(os.path.abspath(__file__))

is_release = False
if "--release" in sys.argv:
    is_release = True
    sys.argv.remove("--release")


setup(name='tycho',
      setup_requires=["vcver", "setuptools-parcels"],
      vcver={"is_release": is_release,
             "path": BASE},
      description='',
      author="zillow-orbital",
      author_email="",
      long_description=open('README.rst').read(),
      packages=find_packages(),
      install_requires=[
          'aiohttp',
          'aiohttp-cors',
          'aiohttp-transmute',
          'motor',
          'pymongo',
          'orbital-core',
      ],
      entry_points={
          'console_scripts': [
              'create_indexes=tycho.scripts.create_indexes:main'
          ]
      },
      include_package_data=True,
      parcels={}
)



import attr
from typing import List, Dict
from .event import ignore_microseconds
from datetime import datetime


def ignore_microseconds_list(value):
    return [ignore_microseconds(d) for d in value]


@attr.s
class EventDB:
    """ This model represents event data format in DB """
    id = attr.ib(type=str)
    time = attr.ib(type=List[datetime], converter=ignore_microseconds_list)
    update_time = attr.ib(type=datetime)
    tags = attr.ib(type=List[str], default=attr.Factory(list))
    detail_urls = attr.ib(type=Dict[str, str], default=attr.Factory(dict))
    description = attr.ib(type=str, default="")

    def asdict(self):
        eventdb_dict = attr.asdict(self)
        eventdb_dict["_id"] = eventdb_dict.pop("id")
        return eventdb_dict



import attr
import pytz
from bson import ObjectId
from datetime import datetime
from uuid import uuid4
from typing import Dict, List
from transmute_core.object_serializers.cattrs_serializer import create_cattrs_converter


CATTRS_CONVERTER = create_cattrs_converter()


def ensure_value_is_not_empty(instance, attribute, value):
    if not value:
        raise ValueError("Values must not be none or empty")
    return value


def ensure_maxlength_100_chars_and_tags_key_val_not_none(instance, attribute, value):
    """
    A helper function to ensure that the values for
    the reserved kays are less than a hundred character limit
    and that the tags do not have any none keys or vals
    """
    if isinstance(value, str):
        if len(value) > 100:
            raise ValueError("Values must not "
                             "exceed length of 100 chars")
    elif isinstance(value, dict):
        for key in value:
            if key is None or key == "None":
                raise ValueError("Tags keys can't be None")
            if value.get(key) is None:
                raise ValueError("Tags values can't be None")
            for elem in value[key]:
                if elem is None:
                    raise ValueError("Tags values can't be None")
                if len(elem) > 100:
                    raise ValueError("Values must not "
                                     "exceed length of 100 chars")
    return value


def ensure_value_is_not_none(instance, attribute, value):
    """
    A helper function to ensure that the values for
    the time is not none
    """
    if value is None:
        raise ValueError("Value can't be None")


def ignore_microseconds(value):
    if value.tzinfo is None:
        utc = pytz.timezone('UTC')
        value = utc.localize(value)
    value = value.astimezone(pytz.utc)
    microsecond = value.microsecond
    new_microsecond = int(microsecond / 1000) * 1000
    return value.replace(microsecond=new_microsecond)


@attr.s
class Event(object):
    """
    This class defines an event with the following reserved fields:
        id,
        source_id,
        parent_id,
        start_time,
        end_time,
        description,
        detail_urls

    and allows other optional attributes to be added depending on the type
    of the event that go inside the tags field.

    An example of an event as a json object is:
    {
       "id": "333f1f77bcf86cd799439011"
       "source_id": "222f1f77bcf86cd799439011",
       "parent_id": "111f1f77bcf86cd799439011",
       "start_time": "12/05/16:10:25:00",
       "end_time": "12/05/16:10:26:00",
       "description": "This is a deploy deployment",
       "detail_urls": {"graphite": "http//graphite"}
       "tags": {
                  "source": ["deploy"],
                  "type": ["deploy/deploy_all"],
                  "author": ["user@example.com", "user2@example.com"],
                  "environment": ["monitor_candidate"],
                  "services": ["tycho"],
                  "status": ["success"]
                }
    }
    """
    # reserved keys
    # assign uuid as id for event if not provided
    id = attr.ib(type=str, default=attr.Factory(lambda: str(uuid4())),
                 validator=[ensure_value_is_not_empty, ensure_maxlength_100_chars_and_tags_key_val_not_none])
    source_id = attr.ib(type=str, default="",
                        validator=[ensure_maxlength_100_chars_and_tags_key_val_not_none])
    parent_id = attr.ib(type=str, default="",
                        validator=[ensure_maxlength_100_chars_and_tags_key_val_not_none])
    start_time = attr.ib(type=datetime, default=attr.Factory(datetime.utcnow),
                         converter=ignore_microseconds, validator=[ensure_value_is_not_none])
    end_time = attr.ib(type=datetime, default=attr.Factory(datetime.utcnow),
                       converter=ignore_microseconds, validator=[ensure_value_is_not_none])
    description = attr.ib(type=str, default="")
    detail_urls = attr.ib(type=Dict[str, str], default=attr.Factory(dict))
    tags = attr.ib(type=Dict[str, List[str]], default=attr.Factory(dict),
                   validator=[ensure_maxlength_100_chars_and_tags_key_val_not_none])

    @staticmethod
    def from_dict(d):
        return CATTRS_CONVERTER.structure(d, Event)

    def to_primitive(self):
        return CATTRS_CONVERTER.unstructure(self)
